/*
 * xmlstack -- elf x86 executable call stack tracing application
 * 5/11/2002
 *
 * Please refer to the README file for credit and distribution information.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "xmlstack.h"

#include "elf32.h"
#include "libdis.h"

void binaryProcessStaticSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx);
void binaryProcessDynamicSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx);
void binaryReadInstructions(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *parent, unsigned char *opcodes, unsigned int len);

Elf32_Sym *binaryLookupSymbolByName(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, const char *name, char *symbolType);
Elf32_Sym *binaryLookupSymbolByAddr(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, unsigned long addr, char *symbolType);
char *binaryLookupSymbolName(ELF32_CTX *ctx, Elf32_Sym *sym);
void binaryTraverseExecutable(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *parent);
void binaryLookupDebugSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *me);
void binaryPrecacheDebugSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx);

void binaryFilterFunctions(XMLSTACK_INSTANCE *xsi);
char binaryFilterFunctionsRecursively(DOM_Element *node, const char **functions, int numFunctions);

int binaryProcess(XMLSTACK_INSTANCE *xsi)
{
	ELF32_CTX ctx;
	Elf32_Sym *mainSymbol = NULL;
	char symType;

	if (!xsi->binaryFile)
	{
		if (xsi->flags & XMLSTACK_FLAG_VERBOSE)
			fprintf(stderr,"binaryProcess(): invalid binary file specified.\n");

		return 0;
	}

	if (xsi->flags & XMLSTACK_FLAG_WRITEASYOUGO)
	{
		if (xsi->xmlFile)
			xsi->writeAsYouGoFd = fopen(xsi->xmlFile, "w");
		else if (!xsi->writeAsYouGoFd)
			xsi->writeAsYouGoFd = stdout;	
	}

	elf32_initialize(&ctx);

	if (elf32_open(&ctx, xsi->binaryFile) != ELF32_SUCCESS)
	{
		if (xsi->flags & XMLSTACK_FLAG_VERBOSE)
			fprintf(stderr,"binaryProcess(): invalid elf file specified. (%s)\n",xsi->binaryFile);

		return 0;
	}

	if (elf32_loadImage(&ctx) != ELF32_SUCCESS)
	{
		if (xsi->flags & XMLSTACK_FLAG_VERBOSE)
			fprintf(stderr,"binaryProcess(): failed to load elf image.\n");

		elf32_close(&ctx);
		return 0;
	}

	if (xsi->flags & XMLSTACK_FLAG_PRECACHEDBGSYMS)
		binaryPrecacheDebugSymbols(xsi, &ctx);

	xsi->rootElement = dom_document_create_element("Program");

	if (!xsi->rootElement)
	{
		elf32_close(&ctx);
		elf32_deinitialize(&ctx);
		return 0;
	}

	if (!(xsi->flags & XMLSTACK_FLAG_NOTEXECUTABLE))
	{
		dom_element_set_attribute(xsi->rootElement, "executable", xsi->binaryFile);	

		mainSymbol = binaryLookupSymbolByName(xsi, &ctx, "main", &symType);

		if (!mainSymbol)
		{
			if (xsi->flags & XMLSTACK_FLAG_VERBOSE)
				fprintf(stderr,"failed to find main symbol, aborting.\n");

			elf32_close(&ctx);
			elf32_deinitialize(&ctx);

			return 0;
		}

		binaryTraverseExecutable(xsi, &ctx, mainSymbol, xsi->rootElement);
	}
	else
	{
		Elf32_Sym *sym;
		int x = 0;

		dom_element_set_name(xsi->rootElement, "Object");
		dom_element_set_attribute(xsi->rootElement, "file", xsi->binaryFile);

		while ((sym = elf32_symbolTableEnum(&ctx, x++)))
			binaryTraverseExecutable(xsi, &ctx, sym, xsi->rootElement);
	}

	if (!(xsi->flags & XMLSTACK_FLAG_WRITEASYOUGO))
	{
		if (xsi->filterFunctions)
			binaryFilterFunctions(xsi);
	
		if (xsi->xmlFile)
		{
			FILE *fd = fopen(xsi->xmlFile,"w");
	
			if (!fd)
				fprintf(stderr,"failed to open '%s'.\n",xsi->xmlFile);
			else
			{
				dom_node_recursive_output_fileptr(xsi->rootElement, NULL, fd);
				fclose(fd);
			}
		}
		else
			dom_node_recursive_output_fileptr(xsi->rootElement, NULL, stdout);
	}
	
	if (xsi->flags & XMLSTACK_FLAG_CLEANUP)
	{
		dom_node_destroy(xsi->rootElement);
	
		if (xsi->flags & XMLSTACK_FLAG_PRECACHEDBGSYMS)
			hashDeinitialize(&xsi->dbgSymHash);
	}

	elf32_close(&ctx);
	elf32_deinitialize(&ctx);

	return 1;
}

Elf32_Sym *binaryLookupSymbolByName(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, const char *name, char *symbolType)
{
	Elf32_Sym *sym;
	int x = 0;

	/* Search static symbols. */

	while ((sym = elf32_symbolTableEnum(ctx, x++)))
	{
		if (ELF32_ST_TYPE(sym->st_info) != STT_FUNC)
			continue;

		if ((elf32_symbolGetName(ctx, sym)) &&
				(!strcmp(elf32_symbolGetName(ctx, sym), name)))
		{
			*symbolType = SYMBOL_TYPE_STATIC;
			break;
		}
	}

	if (sym)
		return sym;

	/* Now search dynamic symbols. */

	x = 0;

	while ((sym = elf32_dynamicSymbolTableEnum(ctx, x++)))
	{
		if (ELF32_ST_TYPE(sym->st_info) != STT_FUNC)
			continue;

		if ((elf32_dynamicSymbolGetName(ctx, sym)) &&
				(!strcmp(elf32_dynamicSymbolGetName(ctx, sym), name)))
		{
			*symbolType = SYMBOL_TYPE_DYNAMIC;
			break;
		}
	}

	return sym;
}

Elf32_Sym *binaryLookupSymbolByAddr(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, unsigned long addr, char *symbolType)
{
	Elf32_Sym *sym;
	int x = 0;

	/* Search static symbols. */

	while ((sym = elf32_symbolTableEnum(ctx, x++)))
	{
		if (ELF32_ST_TYPE(sym->st_info) != STT_FUNC)
			continue;

		if (sym->st_value == addr)
		{
			*symbolType = SYMBOL_TYPE_STATIC;
			break;
		}
	}

	if (sym)
		return sym;

	/* Now search dynamic symbols. */

	x = 0;

	while ((sym = elf32_dynamicSymbolTableEnum(ctx, x++)))
	{
		if (ELF32_ST_TYPE(sym->st_info) != STT_FUNC)
			continue;

		if (sym->st_value == addr)
		{
			*symbolType = SYMBOL_TYPE_DYNAMIC;
			break;
		}
	}

	return sym;

}

void binaryTraverseExecutable(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *parent)
{
	Elf32_Shdr *section = elf32_sectionHeaderEnum(ctx, sym->st_shndx);
	unsigned char *opcodes;
	DOM_Element *me, *parentParent = parent, *child;
	char *symName = (char *)elf32_symbolGetName(ctx, sym), *value;
	char symAddrStr[16];


	me = dom_element_create(parent);
	dom_element_set_name(me, "Function");

	if (symName)
		dom_element_set_attribute(me, "name", symName);
	else
	{
		snprintf(symAddrStr, sizeof(symAddrStr) - 1, "%.8x", sym->st_value);
		dom_element_set_attribute(me, "name", symAddrStr);
		symName = symAddrStr;
	}

	if (xsi->flags & XMLSTACK_FLAG_DBGSYMS)
		binaryLookupDebugSymbols(xsi, ctx, sym, me);

	while ((parentParent = dom_node_get_parent(parentParent)))
	{
		if ((dom_element_get_attribute(parentParent,"name")) && 
				(!strcmp(dom_element_get_attribute(parentParent,"name"), symName)))
			break;
	}

	if (parentParent)
	{
		/* The parent of the parent is us which means we have a referential loop. */
		dom_element_set_attribute(me, "referential", "1");
		dom_element_set_attribute(me, "referential-parent", dom_element_get_attribute(parentParent,"name"));
		return;
	}

	if (xsi->flags & XMLSTACK_FLAG_DONTREPEATSTK)
	{
		for (child = dom_node_get_first_child(parent);
				child;
				child = dom_node_get_next_sibling(child))
		{
			/*
			 * Skip ourselves.
			 */
			if (child == me)
				continue;

			/* If this child already exists inside this parent, skip. */
			if ((dom_element_get_attribute(child, "name")) && (!strcmp(dom_element_get_attribute(child, "name"), symName)))
				return;
		}
	}

	if ((!section) || 					/* invalid section  */
			(!sym->st_value) || 			/* invalid sym addr */
			(!section->sh_addr) || 		/* invalid section addr */
			(section->sh_type != SHT_PROGBITS) ||		/* not in the progbits section */
			(!sym->st_size))				/* no size set			*/
		return;			

	if (lseek(ctx->objectFd, section->sh_offset + (sym->st_value - section->sh_addr), SEEK_SET) < 0)
	{
		fprintf(stderr,"STATICS: failed lseek for symbol '%s'.\n",elf32_symbolGetName(ctx, sym));
		return;
	}

	if (!(opcodes = (char *)malloc(sym->st_size)))
	{
		fprintf(stderr,"STATICS: failed to malloc size '%d'.\n",sym->st_size);
		return;
	}

	if (read(ctx->objectFd, opcodes, sym->st_size) <= 0)
	{
		fprintf(stderr,"STATICS: failed to read '%d' bytes.\n",sym->st_size);
		free(opcodes);
		return;
	}

	if (xsi->flags & XMLSTACK_FLAG_WRITEASYOUGO)
	{
		fprintf(xsi->writeAsYouGoFd, "<Function name=\"%s\"", symName);
		
		if (dom_element_get_attribute(me, "referential"))
			fprintf(xsi->writeAsYouGoFd, " referential=\"1\"");
		if (dom_element_get_attribute(me, "recursive"))
			fprintf(xsi->writeAsYouGoFd, " recursive=\"1\"");
		if ((value = dom_element_get_attribute(me, "referential-parent")))
			fprintf(xsi->writeAsYouGoFd, " referential-parent=\"%s\"", value);
		if ((value = dom_element_get_attribute(me, "srcpath")))
			fprintf(xsi->writeAsYouGoFd, " srcpath=\"%s\"", value);
		if ((value = dom_element_get_attribute(me, "srcline")))
			fprintf(xsi->writeAsYouGoFd, " srcline=\"%s\"", value);

		fprintf(xsi->writeAsYouGoFd, ">");
	}

	binaryReadInstructions(xsi, ctx, sym, me, opcodes, sym->st_size);

	free(opcodes);

	if (xsi->flags & XMLSTACK_FLAG_WRITEASYOUGO)
	{
		fprintf(xsi->writeAsYouGoFd, "</Function>");

		dom_node_destroy(me);
	}
}

void binaryReadInstructions(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *parent, unsigned char *opcodes, unsigned int len)
{
	struct instr i;
	unsigned int x = 0, size;
	char symType;

	while (x < len)
	{
		size = disassemble_address(opcodes + x, &i);

		if (size) 
		{
			x += size;

			if ((i.mnemonic[0] == 'c') && 
					(i.mnemonic[1] == 'a') && 
					(i.mnemonic[2] == 'l') && 
					(i.mnemonic[3] == 'l'))
			{
				Elf32_Sym *childSymbol;
				unsigned long addr = 0;
				long offset = 0;

				if (i.dest[0] == '+')			
					offset = strtol(i.dest+1,NULL,16);
				else
				{
					unsigned long t = strtoul(i.dest+1,NULL,16);
					offset = (-((~t) + 1));
				}

				addr = (sym->st_value+x) + offset;

				childSymbol = binaryLookupSymbolByAddr(xsi, ctx, addr, &symType);

				if (childSymbol)
				{
					if (childSymbol == sym)
					{
						if (!dom_element_get_attribute(parent, "recursive"))
							dom_element_set_attribute(parent, "recursive", "1");
					}
					else if (childSymbol)
						binaryTraverseExecutable(xsi, ctx, childSymbol, parent); 
				}
			}	
		}
		else
			x++;	
	}
}

void binaryLookupDebugSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx, Elf32_Sym *sym, DOM_Element *me)
{
	Elf32_Symbol_Debug dbgInfo, *cachedInfo = NULL;

	if (xsi->flags & XMLSTACK_FLAG_PRECACHEDBGSYMS)
		cachedInfo = (Elf32_Symbol_Debug *)hashIntegerKeyGetValue(&xsi->dbgSymHash, (int)sym);	
	else
		elf32_debugSymbolGetSymbolInfo(ctx, sym, &dbgInfo);

	if (!cachedInfo)
		cachedInfo = &dbgInfo;

	if (cachedInfo->path[0])
		dom_element_set_attribute(me, "srcpath", cachedInfo->path);
	if (cachedInfo->line[0])
		dom_element_set_attribute(me, "srcline", cachedInfo->line);
}

void binaryPrecacheDebugSymbols(XMLSTACK_INSTANCE *xsi, ELF32_CTX *ctx)
{
	Elf32_Sym *sym;
	int x = 0;

	hashInitialize(&xsi->dbgSymHash, HASH_FLAG_ISSTATIC | HASH_FLAG_FREEVALUE, 0);

	while ((sym = elf32_symbolTableEnum(ctx, x++)))
	{
		Elf32_Symbol_Debug *newInfo = (Elf32_Symbol_Debug *)malloc(sizeof(Elf32_Symbol_Debug));

		memset(newInfo, 0, sizeof(Elf32_Symbol_Debug));

		elf32_debugSymbolGetSymbolInfo(ctx, sym, newInfo);

		hashIntegerKeyAdd(&xsi->dbgSymHash, (int)sym, newInfo);	
	}
}

void binaryFilterFunctions(XMLSTACK_INSTANCE *xsi)
{
	char *functions[128], currFunction[128];
	int currFunctionPos = 0, len = strlen(xsi->filterFunctions), x;
	int numFunctions = 0;
	
	for (x = 0; x <= len; x++)
	{
		if ((x == len) ||
			 (xsi->filterFunctions[x] == ','))
		{
			currFunction[currFunctionPos] = 0;
			functions[numFunctions++] = strdup(currFunction);

			if (!functions[numFunctions-1])
				return;

			memset(currFunction,0,currFunctionPos);
			currFunctionPos = 0;
		}
		else if (currFunctionPos+1 < sizeof(currFunction))
			currFunction[currFunctionPos++] = xsi->filterFunctions[x];
	}

	if (numFunctions > 0)
		binaryFilterFunctionsRecursively(xsi->rootElement, (const char **)functions, numFunctions);

	for (x = 0; x < numFunctions; x++)
		free(functions[x]);
}

char binaryFilterFunctionsRecursively(DOM_Element *node, const char **functions, int numFunctions)
{
	DOM_Element *child, *next;
	char *functionName = dom_element_get_attribute(node,"name");
	int x = 0;
	char found = 0;

	if (dom_node_get_parent(node) && functionName)
	{
		for (x = 0; x < numFunctions; x++)
		{
			if (!strcmp(functionName, functions[x]))
			{
				DOM_Node *child;
				
				while ((child = dom_node_get_first_child(node)))
					dom_node_destroy(child);

				return 1;
			}
		}
	}		

	child = dom_node_get_first_child(node);

	while (child)
	{
		next = dom_node_get_next_sibling(child);

		if (!binaryFilterFunctionsRecursively(child, functions, numFunctions))
			dom_node_destroy(child);
		else 
			found = 1;

		child = next;
	}

	return found;
}
